"""Browser profile models."""

import json
from enum import Enum
from typing import Union

from pydantic import BaseModel, DirectoryPath, Field

from pybas_automation import STORAGE_SUBDIR, default_model_config
from pybas_automation.bas_actions.browser.proxy.models import BasActionBrowserProxy
from pybas_automation.browser_profile.settings import _proxy_filename, _user_data_dir_default_factory


class BrowserProfileProxyEnum(str, Enum):
    """Enum class for possible proxy types."""

    HTTP = "http"
    SOCKS5 = "socks5"


class BrowserProfileProxy(BaseModel):
    """Defines the proxy settings for a browser profile."""

    model_config = default_model_config

    hostname: str = Field(default=str)
    port: int = Field(default=3128, ge=1, le=65535)
    type: BrowserProfileProxyEnum = Field(default=BrowserProfileProxyEnum.SOCKS5)
    login: str = Field(default="", max_length=200)
    password: str = Field(default="", max_length=200)


class BrowserProfile(BaseModel):
    """Represents a browser profile with customizable settings."""

    model_config = default_model_config

    profile_dir: DirectoryPath = Field(default_factory=_user_data_dir_default_factory)
    fingerprint_raw: Union[str, None] = Field(default=None)
    proxy: Union[BrowserProfileProxy, None] = Field(default=None)

    def save_proxy_to_profile(self) -> bool:
        """
        Save the proxy to the profile directory.

        :return: True if the proxy was saved successfully, False otherwise.
        """

        if self.proxy is None:
            return False

        is_http: bool = self.proxy.type == BrowserProfileProxyEnum.HTTP

        bas_proxy = BasActionBrowserProxy(
            server=self.proxy.hostname,
            port=f"{self.proxy.port}",  # type: ignore
            is_http=is_http,  # type: ignore
            name=self.proxy.login,
            password=self.proxy.password,
        )

        sub_dir = self.profile_dir.joinpath(STORAGE_SUBDIR)
        sub_dir.mkdir(parents=True, exist_ok=True)

        proxy_filename = sub_dir.joinpath(_proxy_filename)
        proxy_filename.open("w", encoding="utf-8").write(json.dumps(bas_proxy.model_dump(by_alias=True)))

        return True



"""Collection of utility functions."""

from .filesystem import create_storage_dir_in_app_data
from .logger import get_logger
from .utils import timing

__all__ = ["create_storage_dir_in_app_data", "get_logger", "timing"]



"""
Functions for getting fingerprints from the fingerprint API.
"""

from urllib.parse import urlencode

import httpx

from pybas_automation.fingerprint.models import BasFingerprintRequest

FINGERPRINT_BASE_URL = "https://fingerprints.bablosoft.com/prepare?version=5"


class FingerprintRequestException(Exception):
    """Raised when a fingerprint request fails."""


def get_fingerprint(
    request_data: BasFingerprintRequest,
) -> str:
    """Get a fingerprint for the given fingerprint key."""

    json_data = dict(request_data.model_dump())

    json_data["tags"] = ",".join(request_data.tags)
    json_data["returnpc"] = "true"

    url = f"{FINGERPRINT_BASE_URL}&{urlencode(json_data)}"

    response = httpx.get(url, timeout=10)
    if response.status_code != 200:
        raise FingerprintRequestException(f"Failed to get fingerprint: {response.text}")

    response_json_data = response.json()
    if not response_json_data.get("valid"):
        raise FingerprintRequestException(f"Failed to get fingerprint: {response_json_data}")

    return response.text.strip()



"""
All actions to work with browser.

These actions  do not require a specific element for use.
For example, load the url, set proxy, make a screenshot, etc. To click on an element or enter text in a specific field,
click on this item and select an action from the menu.
"""



"""
Settings for the task module.
"""


from pydantic import DirectoryPath, FilePath

_storage_dir = DirectoryPath("PyBASProfilesTasks")

_task_filename = FilePath("tasks.json")
_filelock_filename = FilePath("tasks.lock")



"""Module for interacting with the BAS fingerprint API."""

from .fingerprint import FingerprintRequestException, get_fingerprint
from .models import BasFingerprintRequest

__all__ = [
    "BasFingerprintRequest",
    "FingerprintRequestException",
    "get_fingerprint",
]



"""Browser profile proxy module."""

from typing import Dict

import httpx

from pybas_automation.browser_profile.models import BrowserProfileProxy


class ExternalIPRequestException(Exception):
    """Raised when an error occurs while requesting the external IP address."""


def get_external_info_ip(profile_proxy: BrowserProfileProxy) -> Dict:
    """Get the external IP address."""

    proxy_str = f"{profile_proxy.hostname}:{profile_proxy.port}"

    if profile_proxy.login and profile_proxy.password:
        proxy_str = f"{profile_proxy.login}:{profile_proxy.password}@{proxy_str}"

    match profile_proxy.type:
        case "socks5":
            proxies = f"socks5://{proxy_str}"
        case "http":
            proxies = f"http://{proxy_str}"
        case _:
            raise ValueError(f"Invalid proxy type: {profile_proxy.type}")
    try:
        response = httpx.get(url="https://lumtest.com/myip.json", proxies=proxies, timeout=10)
    except Exception as exc:
        raise ExternalIPRequestException("Failed to get external IP.") from exc

    if response.status_code != 200:
        raise ExternalIPRequestException(f"Failed to get external IP: {response.text}")

    try:
        response_json_data = response.json()
    except Exception as exc:
        raise ExternalIPRequestException("Failed to get external IP.") from exc

    return dict(response_json_data)



"""
Utility functions for the application.
"""

from functools import wraps
from time import time

from pybas_automation.utils import get_logger

logger = get_logger()


def timing(f):  # type: ignore
    """Time functions for debugging purposes."""

    @wraps(f)
    def wrap(*args, **kw):  # type: ignore
        ts = time()
        result = f(*args, **kw)
        te = time()
        # pylint: disable=logging-fstring-interpolation
        logger.debug(f"func:{f.__name__} args:[{args}, {kw}] took: {te - ts:.4f} sec")

        return result

    return wrap



"""Browser / Proxy models."""

from typing import Optional

from pydantic import BaseModel, Field, field_validator

from pybas_automation import default_model_config


class BasActionBrowserProxy(BaseModel):
    """BasActionBrowserProxy is used to specify a proxy for a browser profile."""

    model_config = default_model_config

    server: str = Field(default="127.0.0.1")
    port: str = Field(default="3128", alias="Port")
    is_http: Optional[bool] = Field(alias="IsHttp", default=False)
    name: Optional[str] = Field(default="")
    password: Optional[str] = Field(default="")

    @field_validator("port")
    @classmethod
    def port_str_must_be_integer(cls, v: str) -> str:
        """Validate that port is an integer and in range 1-65535."""
        try:
            v_int = int(v)
        except ValueError as exc:
            raise ValueError(f"must be integer, got: {v}") from exc

        if v_int < 1 or v_int > 65535:
            raise ValueError(f"must be in range 1..65535, got: {v}")

        return v



"""BasTask models."""

from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from pybas_automation import default_model_config
from pybas_automation.bas_actions.browser.browser_settings.models import BasActionBrowserSettings


class BasTask(BaseModel):
    """A task contains all the information needed to run a task via BAS GUI."""

    model_config = default_model_config

    # unique task id
    task_id: UUID = Field(default_factory=uuid4)

    browser_settings: BasActionBrowserSettings = Field(default_factory=BasActionBrowserSettings)



"""
Filesystem utilities.
"""

import os

from pydantic import DirectoryPath


def create_storage_dir_in_app_data(storage_dir: DirectoryPath) -> DirectoryPath:
    """
    Create a storage directory in the local app data directory.

    :param storage_dir: The name of the storage directory.
    :return: Directory path to the created storage directory in the local app data directory.
    """

    _env_path = os.getenv("LOCALAPPDATA")
    if not _env_path:
        raise ValueError("Environment variable LOCALAPPDATA is not set.")

    local_app_data_path = DirectoryPath(_env_path)

    if not local_app_data_path.is_dir() or not local_app_data_path.exists():
        raise ValueError("Cannot find local app data path.")

    storage_dir = local_app_data_path.joinpath(storage_dir)
    if not storage_dir.exists():
        os.mkdir(storage_dir)

    return DirectoryPath(storage_dir)



"""Browser / Browser Settings models."""

from enum import Enum
from typing import List, Union

from pydantic import BaseModel, DirectoryPath, Field

from pybas_automation import default_model_config
from pybas_automation.bas_actions.browser.browser_settings import browser_command_line_factory
from pybas_automation.bas_actions.browser.proxy.models import BasActionBrowserProxy
from pybas_automation.bas_actions.fingerprint_switcher.apply_fingerprint.models import BasActionApplyFingerprintModel


class WidevineEnum(str, Enum):
    """Widevine settings model."""

    ENABLE = "enable"
    DISABLE = "disable"
    EMPTY = ""


class SafeBrowsingEnum(str, Enum):
    """Safe Browsing settings model."""

    ENABLE = "enable"
    DISABLE = "disable"
    EMPTY = ""


class ChromeComponentsEnum(str, Enum):
    """Chrome components settings model."""

    ENABLE = "enable"
    DISABLE = "disable"
    EMPTY = ""


class BasActionBrowserSettingsComponents(BaseModel):
    """Bas Browser components settings model."""

    model_config = default_model_config

    # Enable or disable Widevine plugin.
    # enable - Use Widevine
    # disable - Don't use Widevine
    # Empty string - Don't change this setting
    widevine: WidevineEnum = Field(default=WidevineEnum.ENABLE)

    # Enable Safe Browsing. It is technology embedded in Chrome, helps to protect users from malicious websites.
    # Its absence can be detected.
    # enable - Enable Safe Browsing
    # disable - Disable Safe Browsing
    # Empty string - Don't change this setting
    safe_browsing: SafeBrowsingEnum = Field(default=SafeBrowsingEnum.ENABLE)

    # Enable Chrome components. Increase profile size, but its absence can be detected.
    # enable - Enable components
    # disable - Disable components
    # Empty string - Don't change this setting
    components: ChromeComponentsEnum = Field(default=ChromeComponentsEnum.ENABLE)


class BasActionBrowserSettingsRendering(BaseModel):
    """Bas Browser rendering settings model."""

    model_config = default_model_config

    # Maximum times browser content can be rendered per one second.
    # The lower this value, the lower the CPU usage will be. Setting it too low may affect site operability.
    # Reducing it below 30 may lead to unpredictable consequences. Minimum value is 10.
    maximum_fps: int = Field(default=30, gt=10, lt=60)


class BasActionBrowserSettingsNetwork(BaseModel):
    """Bas Browser network settings model."""

    # Unlike HTTP, QUIC protocol is build on top of UDP. Not all proxies supports UDP.
    # It means that enabling QUIC can cause problems when working with certain proxies.
    # It is recommended to enable this option only if sure, that your proxy supports it. Disabled by default.
    # enable - Enable QUIC
    # disable - Disable QUIC
    enable_qiuc_protocol: bool = Field(default=True)


class BasActionBrowserSettingsProfile(BaseModel):
    """Bas Browser profile settings model."""

    model_config = default_model_config

    # String with profile folder. Slash type is not important. If folder doesn't exist, it will be created. If folder
    # already exists, BAS will use it as profile and restore all data from it like cookies, localstorage,
    # etc. By default, browser stores all profile data in temporary folder, you can use "temporary" keyword to switch
    # to new temporary profile. Empty string - Don't change temporary - Switch to new temporary profile
    profile_folder_path: DirectoryPath = Field(default="")

    # In case if profile folder already exists and has fingerprint data,
    # tells BAS to apply fingerprint used latest for that profile.
    always_load_fingerprint_from_profile_folder: bool = Field(default=False)

    # In case if profile folder already exists and has proxy data,
    # tells BAS to apply proxy used latest for that profile.
    always_load_proxy_from_profile_folder: bool = Field(default=False)


class BasActionBrowserSettings(BaseModel):
    """Browser / Browser Settings model."""

    model_config = default_model_config

    components: BasActionBrowserSettingsComponents = Field(default_factory=BasActionBrowserSettingsComponents)
    network: BasActionBrowserSettingsNetwork = Field(default_factory=BasActionBrowserSettingsNetwork)
    rendering: BasActionBrowserSettingsRendering = Field(default_factory=BasActionBrowserSettingsRendering)

    # Changes browser version for current thread.
    # This setting will restart browser and therefore reset all settings, so it is better to use it when thread starts.
    browser_version: str = Field(default="default")

    # Chromium command line arguments.
    command_line: Union[List[str]] = Field(default_factory=browser_command_line_factory)

    # Profile folder path.
    profile: BasActionBrowserSettingsProfile = Field(default_factory=BasActionBrowserSettingsProfile)

    # Proxy settings.
    proxy: Union[BasActionBrowserProxy, None] = Field(default=None)

    # Browser fingerprint. Fingerprint switcher -> Apply fingerprint
    fingerprint: BasActionApplyFingerprintModel = Field(default_factory=BasActionApplyFingerprintModel)



"""Task storage module. This module is responsible for storing tasks to disk and loading tasks from disk into memory."""

import json
import os
from enum import Enum
from typing import Literal, Set, Union
from uuid import UUID

import filelock
from pydantic import DirectoryPath, FilePath

from pybas_automation.task.models import BasTask
from pybas_automation.task.settings import _filelock_filename, _storage_dir, _task_filename
from pybas_automation.utils import create_storage_dir_in_app_data, get_logger

logger = get_logger()


class TaskStorageModeEnum(str, Enum):
    """Task storage is used to specify the mode to open the tasks file in."""

    READ = "r"
    READ_WRITE = "rw"


class TaskDuplicateError(Exception):
    """Raised when a task already exists in the storage."""


class TaskStorage:
    """TaskStorage is responsible for storing tasks to disk and loading tasks from disk into memory."""

    storage_dir: DirectoryPath
    mode: TaskStorageModeEnum = TaskStorageModeEnum.READ
    task_file_path: FilePath

    _tasks: Union[list[BasTask], None] = None
    _tasks_unique_id: Set[UUID]
    _lock: filelock.FileLock

    def __init__(
        self,
        storage_dir: Union[None, DirectoryPath] = None,
        task_filename: Union[None, FilePath] = None,
        mode: Union[TaskStorageModeEnum, None] = None,
    ) -> None:
        """
        Initialize TaskStorage. If the storage_dir is not provided, the default storage directory will be used.

        :returns: None

        :param storage_dir: The directory to store the tasks file.
        :param task_filename: The filename of the tasks file.
        :param mode: The mode to open the tasks file in. Defaults to read-only.

        :raises ValueError: If the storage_dir is not a directory. If the mode is not a valid value.
        """

        if storage_dir is None:
            self.storage_dir = create_storage_dir_in_app_data(storage_dir=_storage_dir)
        else:
            if not os.path.isdir(storage_dir):
                raise ValueError(f"storage_dir is not a directory: {storage_dir}")
            self.storage_dir = DirectoryPath(storage_dir)

        if task_filename is None:
            self.task_file_path = FilePath(os.path.join(self.storage_dir, _task_filename))
        else:
            task_filename = FilePath(task_filename)
            if task_filename.parent.__str__() != ".":
                raise ValueError(f"task_filename is not a relative path: {task_filename}")

            self.task_file_path = FilePath(os.path.join(self.storage_dir, task_filename))

        # Set the mode of the task storage
        match mode:
            case None:
                self.mode = TaskStorageModeEnum.READ
            case TaskStorageModeEnum.READ:
                self.mode = TaskStorageModeEnum.READ
            case TaskStorageModeEnum.READ_WRITE:
                self.mode = TaskStorageModeEnum.READ_WRITE
            case _:
                raise ValueError(f"mode is not a valid value: {mode}")

        self._tasks_unique_id = set()
        self._lock = filelock.FileLock(os.path.join(self.storage_dir, _filelock_filename))

        self.load_all()

    def __repr__(self) -> str:
        """Return a string representation of the TaskStorage."""
        return f"<TaskStorage storage_dir={self.storage_dir} task_file_path={self.task_file_path} mode={self.mode}>"

    def clear(self) -> bool:
        """
        Clear all tasks from the storage. This will also delete the tasks file and clear the tasks in memory.

        :return: True if the tasks were cleared, False otherwise.

        :raises ValueError: If the task storage is in read-only mode.
        """
        if self.mode == TaskStorageModeEnum.READ:
            raise ValueError("Cannot clear tasks in read mode.")
        if self._lock is None:
            raise ValueError("Lock is not initialized.")

        with self._lock:
            self._tasks = None
            self._tasks_unique_id = set()
            if os.path.exists(self.task_file_path):
                self.task_file_path.unlink()
                return True

        return False

    def save(self, task: BasTask) -> None:
        """
        Save a task to the storage.

        :return:  None

        :param task: The task to save.

        :raises ValueError: If the task storage is in read-only mode or if the task already exists.
        """

        if self.mode == TaskStorageModeEnum.READ:
            raise ValueError("Cannot store tasks in read mode.")
        if self._lock is None:
            raise ValueError("Lock is not initialized.")

        with self._lock:
            if self._tasks is None:
                self._tasks = []
            if task.task_id in self._tasks_unique_id:
                raise TaskDuplicateError(f"Task with id {task.task_id} already exists.")

            self._tasks.append(task)
            self._tasks_unique_id.add(task.task_id)

            _tasks = [t.model_dump(mode="json") for t in self._tasks]

            with self.task_file_path.open(mode="w", encoding="utf-8") as f:
                json.dump(_tasks, f, indent=4)

    def save_all(self) -> bool:
        """
        Save all tasks to the storage.

        :return: True if the tasks were saved, False otherwise.

        :raises ValueError: If the task storage is in read-only mode.
        """
        if self.mode == TaskStorageModeEnum.READ:
            raise ValueError("Cannot store tasks in read mode.")

        if self._lock is False:
            raise ValueError("Lock is not initialized.")

        if self._tasks is None:
            raise ValueError("No tasks to save.")

        if self._lock is None:
            raise ValueError("Lock is not initialized.")

        with self._lock:
            with self.task_file_path.open(mode="w", encoding="utf-8") as f:
                json.dump([t.model_dump(mode="json") for t in self._tasks], f, indent=4)

        return True

    def get(self, task_id: UUID) -> Union[BasTask, Literal[False]]:
        """
        Get a task from the storage.

        :param task_id: The task id to get.

        :return: The task if it exists, False otherwise.
        """
        if self._tasks is None:
            return False

        for task in self._tasks:
            if task.task_id == task_id or str(task.task_id) == task_id:
                return task

        return False

    def get_all(self) -> Union[list[BasTask], None]:
        """
        Get all tasks from the storage.

        :return: A list of tasks if they exist, None otherwise.
        """
        if self._tasks is None:
            return None
        return self._tasks

    def count(self) -> int:
        """
        Get the number of tasks in the storage.

        :return:int The number of tasks in the storage.
        """
        if self._tasks is None:
            return 0
        return len(self._tasks)

    def load_all(self) -> bool:
        """
        Load all tasks from the storage into memory.

        :return: True if the tasks were loaded, False otherwise.

        :raises ValueError: If the task storage is in read-only mode.
        """

        # Check if the task file exists.
        if not os.path.exists(self.task_file_path):
            return False

        # Ensure the lock has been initialized.
        if self._lock is None:
            raise ValueError("Lock is not initialized.")

        with self._lock:  # Acquire the lock.
            with self.task_file_path.open(mode="r", encoding="utf-8") as f:
                tasks_from_file = json.load(f)

            # Clear existing tasks in memory.
            self._tasks = []
            self._tasks_unique_id = set()

            # Populate tasks from the file into memory.
            for task_data in tasks_from_file:
                task = BasTask(**task_data)
                self._tasks.append(task)
                self._tasks_unique_id.add(task.task_id)

        return True



"""Task module for interacting with BAS actions."""

from .models import BasTask
from .storage import TaskDuplicateError, TaskStorage, TaskStorageModeEnum

__all__ = ["BasTask", "TaskDuplicateError", "TaskStorage", "TaskStorageModeEnum"]



"""Browser Processes module."""

import json
from typing import Union

import httpx
import websockets

from pybas_automation.utils import get_logger, timing

logger = get_logger()


class BrowserRemoteDebuggingPortNotAvailableError(Exception):
    """Raised when the remote debugging port is not available."""


def _url_to_ws_endpoint(endpoint_url: str) -> str:
    """Get the websocket url from the http endpoint url."""
    if endpoint_url.startswith("ws"):
        return endpoint_url

    logger.debug("<ws preparing> retrieving websocket url from %s", endpoint_url)

    http_url = endpoint_url if endpoint_url.endswith("/") else f"{endpoint_url}/"
    http_url += "json/version/"
    try:
        response = httpx.get(http_url)
    except httpx.ConnectError as exc:
        raise BrowserRemoteDebuggingPortNotAvailableError(
            f"Cannot connect to {http_url}.\n" "This does not look like a DevTools server, try connecting via ws://."
        ) from exc

    if response.status_code != 200:
        raise ValueError(
            f"Unexpected status {response.status_code} when connecting to {http_url}.\n"
            "This does not look like a DevTools server, try connecting via ws://."
        )

    json_data = json.loads(response.text)
    logger.debug("<ws preparing> response: %s", json_data)

    return str(json_data["webSocketDebuggerUrl"])


class BrowserRemote:
    """BrowserProcess is responsible for finding the browser process for a given profile folder path."""

    remote_debugging_port: int
    ws_endpoint: Union[str, None]
    browser_version: Union[str, None]

    def __init__(self, remote_debugging_port: int):
        """
        BrowserProcess instance.

        :param remote_debugging_port: CDP remote debugging port.
        """

        self.remote_debugging_port = remote_debugging_port
        self.ws_endpoint = None
        self.browser_version = None

    def __repr__(self) -> str:
        return f"<BrowserProcess remote_debugging_port={self.remote_debugging_port} ws_endpoint={self.ws_endpoint}>"

    @timing
    def find_ws(self) -> bool:
        """
        Find the websocket endpoint for the browser remote debugging url.

        :return: True if found.

        :raises BrowserRemoteDebuggingPortNotAvailableError: If the remote debugging port is not available.
        """

        if not self._get_ws_endpoint():
            return False

        return True

    def _get_ws_endpoint(self) -> Union[str, bool]:
        if not self.remote_debugging_port:
            return False

        url = f"http://localhost:{self.remote_debugging_port}"
        ws_url = _url_to_ws_endpoint(url)
        self.ws_endpoint = ws_url

        return True

    async def _get_browser_version(self, ws_url: str) -> dict:
        async with websockets.connect(ws_url) as ws:  # type: ignore
            # Send a command to get browser version
            payload = json.dumps({"id": 1, "method": "Browser.getVersion"})
            await ws.send(payload)
            response = await ws.recv()
            return dict(json.loads(response))



"""Browser / Browser Settings."""

from typing import List


def browser_command_line_factory() -> List[str]:
    """Browser defaults command line arguments."""

    return [
        "--disk-cache-size=104857600",  # 100 Mb
        "--disable-gpu-program-cache",
        "--disable-gpu-shader-disk-cache",
        "--disable-features=GpuProcessHighPriorityWin,GpuUseDisplayThreadPriority",
        "--lang=en",
    ]



"""
Return a logger named based on the caller's full module path.
"""

import inspect
import logging


def get_logger() -> logging.Logger:
    """
    Returns a logger named based on the caller's full module path.
    Format: "[<full-module-path>]"
    """

    # Get the module name of the caller.
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])

    if module is None:
        raise ValueError("Could not determine the caller's module.")

    logger_name = f"[{module.__name__}]"

    logger = logging.getLogger(logger_name)
    # Optional: Set the logging level and formatter, or any other logger configuration you want here.

    return logger



"""
Settings for the browser profile.
"""

import tempfile

from pydantic import DirectoryPath, FilePath

from pybas_automation.utils import create_storage_dir_in_app_data

_storage_dir = DirectoryPath("PyBASProfiles")
_fingerprint_raw_filename = FilePath("fingerprint_raw.json")
_proxy_filename = FilePath("proxy.json")

_filelock_filename = FilePath("tasks.lock")


def _user_data_dir_default_factory() -> FilePath:
    """Return the default user data directory."""

    profiles_dir = create_storage_dir_in_app_data(storage_dir=_storage_dir)

    if not profiles_dir.exists():
        profiles_dir.mkdir()

    profile_name = tempfile.mkdtemp(dir=profiles_dir)
    return FilePath(profile_name)



"""Models for the BAS fingerprint API."""

from typing import List

from pydantic import BaseModel, Field

from pybas_automation import default_model_config


def tags_default_factory() -> List[str]:
    """Return the default tags for a fingerprint request."""

    return ["Microsoft Windows", "Chrome"]


class BasFingerprintRequest(BaseModel):
    """BasFingerprintRequest is used to request a fingerprint from the BAS fingerprint API."""

    model_config = default_model_config

    key: str = Field(min_length=64, max_length=64)
    tags: List[str] = Field(default_factory=tags_default_factory)
    min_browser_version: int = Field(default=117, ge=117)
    min_width: int = Field(default=1366)
    min_height: int = Field(default=768)
    max_width: int = Field(default=1920)
    max_height: int = Field(default=1080)



"""
Browser processes module.
"""

from .browser_remote import BrowserRemote, BrowserRemoteDebuggingPortNotAvailableError

__all__ = [
    "BrowserRemote",
    "BrowserRemoteDebuggingPortNotAvailableError",
]



"""Default configuration for models."""
from pydantic import ConfigDict

from .settings import STORAGE_SUBDIR

default_model_config = ConfigDict(populate_by_name=True, extra="forbid")

__all__ = ["default_model_config", "STORAGE_SUBDIR"]



"""
Browser storage module.

This module is responsible for storing browser profiles to disk and  loading metadata of browser profiles from disk
into memory.
"""

import os
import tempfile
from typing import List, Union

import filelock
from pydantic import DirectoryPath

from pybas_automation import STORAGE_SUBDIR
from pybas_automation.browser_profile import BrowserProfile
from pybas_automation.browser_profile.settings import _filelock_filename, _fingerprint_raw_filename, _storage_dir
from pybas_automation.fingerprint import BasFingerprintRequest, get_fingerprint
from pybas_automation.utils import create_storage_dir_in_app_data, get_logger

logger = get_logger()


class BrowserProfileStorageExistsError(Exception):
    """Raised when a browser profile already exists in the storage."""


class FingerprintError(Exception):
    """Raised when a fingerprint key is empty."""


class BrowserProfileStorage:
    """Handles the storage and retrieval of browser profiles."""

    storage_dir: DirectoryPath
    fingerprint_key: Union[str, None]

    _profiles: Union[list[BrowserProfile], None] = None
    _lock: filelock.FileLock

    def __init__(
        self, storage_dir: Union[DirectoryPath, None] = None, fingerprint_key: Union[str, None] = None
    ) -> None:
        """
        Initialize BrowserStorage.

        :param storage_dir: The directory to store the browser profiles.
        :param fingerprint_key: Your personal fingerprint key of FingerprintSwitcher.

        :raises ValueError: If the storage_dir is not a directory.
        """

        if storage_dir is None:
            self.storage_dir = create_storage_dir_in_app_data(storage_dir=_storage_dir)
        else:
            if not os.path.isdir(storage_dir):
                raise ValueError(f"storage_dir is not a directory: {storage_dir}")
            self.storage_dir = DirectoryPath(storage_dir)

        self.fingerprint_key = fingerprint_key
        self._lock = filelock.FileLock(os.path.join(self.storage_dir, _filelock_filename))

    def count(self) -> int:
        """
        Count the number of browser profiles in the storage.

        :return: The number of browser profiles in the storage.
        """

        return len(os.listdir(self.storage_dir))

    def new(self, profile_name: Union[str, None] = None, fingerprint_raw: Union[str, None] = None) -> BrowserProfile:
        """
        Create a new browser profile.

        :param profile_name: The name of the browser profile.
        :param fingerprint_raw: The fingerprint raw string.

        :return: BrowserProfile instance.

        :raises FingerprintKeyEmptyError: If the fingerprint key is empty.
        """

        if self.fingerprint_key is None and fingerprint_raw is None:
            raise FingerprintError("fingerprint_key is required.")

        if fingerprint_raw is not None and self.fingerprint_key is not None:
            raise FingerprintError("fingerprint_key and fingerprint_raw cannot be used together.")

        if profile_name is None:
            profile_dir = DirectoryPath(tempfile.mkdtemp(dir=str(self.storage_dir)))
        else:
            profile_dir = self.storage_dir.joinpath(profile_name)
            if profile_dir.exists():
                raise BrowserProfileStorageExistsError(f"Browser profile already exists: {profile_dir}")
            profile_dir.mkdir(parents=False)

        browser_profile = BrowserProfile(profile_dir=profile_dir)

        if fingerprint_raw is None:
            if self.fingerprint_key is None:  # is this dead code?
                raise FingerprintError("fingerprint_key is required.")

            request_data = BasFingerprintRequest(key=self.fingerprint_key)
            fingerprint_raw = get_fingerprint(request_data)

        browser_profile.fingerprint_raw = fingerprint_raw

        self.save(browser_profile=browser_profile)

        return browser_profile

    def save(self, browser_profile: BrowserProfile) -> None:
        """
        Save the browser profile to disk.

        :param browser_profile: BrowserProfile instance.
        :return: None.
        """

        sub_dir = browser_profile.profile_dir.joinpath(STORAGE_SUBDIR)
        sub_dir.mkdir(parents=True, exist_ok=True)

        fingerprint_filename = sub_dir.joinpath(_fingerprint_raw_filename)

        if browser_profile.fingerprint_raw is not None:
            fingerprint_filename.open("w", encoding="utf-8").write(browser_profile.fingerprint_raw)

    def load(self, profile_name: str) -> BrowserProfile:
        """
        Load a browser profile from disk.

        :param profile_name: The name of the browser profile.
        :return: BrowserProfile instance.
        """
        raise NotImplementedError()

    def load_all(self) -> List[BrowserProfile]:
        """
        Load all browser profiles from disk.

        :return: List[BrowserProfile].
        """
        if self._profiles is None:
            self._profiles = []

        for profile_name in os.listdir(self.storage_dir):
            profile_dir = self.storage_dir.joinpath(profile_name)
            browser_profile = BrowserProfile(profile_dir=profile_dir)

            sub_dir = profile_dir.joinpath(STORAGE_SUBDIR)

            fingerprint_filename = sub_dir.joinpath(_fingerprint_raw_filename)
            if fingerprint_filename.exists():
                fingerprint_raw = fingerprint_filename.open("r", encoding="utf-8").read()
                browser_profile.fingerprint_raw = fingerprint_raw

            self._profiles.append(browser_profile)

        return self._profiles



"""FingerprintSwitcher / Apply fingerprint models."""

from pydantic import BaseModel, Field

from pybas_automation import default_model_config


class BasActionApplyFingerprintModel(BaseModel):
    """Fingerprint switcher -> Apply fingerprint."""

    model_config = default_model_config

    # If these settings is set to true, canvas will be enabled and noise will be added to all data returned from canvas.
    safe_canvas: bool = Field(default=True)

    # If these settings is set to true, PerfectCanvas replacement will be enabled. Fingerprint must contain
    # PerfectCanvas data in order to make it work. See "Get fingerprint" action for explanation.
    use_perfect_canvas: bool = Field(default=True)

    # If these settings is set to true, WebGL will be enabled, noise will be added to WebGL canvas and
    # your hardware properties, like video card vendor and renderer, will be changed
    safe_webgl: bool = Field(default=True)

    # If these settings is set to true, WebGL will be enabled, noise will be added to WebGL canvas and your hardware
    # properties, like video card vendor and renderer, will be changed
    safe_audio: bool = Field(default=True)

    # If these settings is set to true battery API will show different values for each thread, this prevents sites for
    # detecting your real identity. In case if device from which fingerprint was obtained doesn't have battery API,
    # 100% charge level will always be returned.
    safe_battery: bool = Field(default=True)

    # By default, browser searches for fonts only in system font folder. This may lead to inconsistencies during
    # fingerprint emulation if target fingerprint has more fonts than local system. Therefore, it is recommended to
    # download font pack with most popular fonts. This setting allows to use font pack if it is installed.
    # More info about font pack and download link can be found here: https://wiki.bablosoft.com/doku.php?id=fontpack
    use_font_pack: bool = Field(default=True)

    # If these settings is set to true, results of API which returns element coordinates will be updated to protect
    # against 'client rects' fingerprinting.
    safe_element_size: bool = Field(default=False)

    # Chrome browser has Sensor API, which allows to read data from devices like accelerometer, gyroscope or others.
    # Data from that devices is available only on mobile platforms. After checking this setting data for that devices
    # will be generated and replaced automatically. Enable this option in order to emulate mobile fingerprints more
    # precisely.
    emulate_sensor_api: bool = Field(default=True)

    # Allows to better emulate devices with higher pixel density. With this setting enabled, emulation will be done
    # in the most natural way. It means that browser will render the page in a bigger resolution, just like on real
    # device. The tradeoff is higher system resources usage, because you need to perform more calculations to render
    # a bigger picture. Javascript settings related to pixel density, for example devicePixelRatio, will be replaced
    # correctly regardless if this setting will be enabled or not.
    emulate_device_scale_factor: bool = Field(default=True)



"""
FingerprintSwitcher is like user agent switcher on steroids.

It changes not only user agent, but a lot of browser internals to make BAS actually look like firefox, chrome, safari,
on desktop or on mobile.
"""



"""
The `browser_profile` module.

This module is responsible for managing browser profiles. It provides functionalities to create, store,
and load browser profiles. The profiles can be customized with different settings like fingerprints and proxies.
"""

from .models import BrowserProfile, BrowserProfileProxy, BrowserProfileProxyEnum
from .storage import BrowserProfileStorage

__all__ = ["BrowserProfile", "BrowserProfileProxy", "BrowserProfileProxyEnum", "BrowserProfileStorage"]
